#lang racket

; Types: CFG = Hash[Exp, Set[Exp]]
;          maps expressions to sets of successors
;        Store = Hash[Var, Set[Lam U Var]]
;          maps variables to sets of lambdas that can
;          be reached from that variable (directly or via a prim-op like car)
;          or variable names representing all possible addresses generated by make-vector or vector
;        Expressions must come from the output of simplify-ae

(define (successors exp store) 
  (define (lookup x) (hash-ref store x set)) 
  (match exp
         [`(let ([,x (quote ,dat)]) ,e0)
          (cons (set e0) store)]
         [`(let ([,x (lambda ,xs ,elam)]) ,e0)
          (cons (set e0) (store-extend store x `(lambda ,xs ,elam)))]
         [`(let ([,x (prim vector ,as ...)]) ,e0)
          (cons (set e0) (store-extend store x (set-union (set x) (foldl set-union (set) (map lookup as)))))]
         [`(let ([,x (apply-prim vector ,args)]) ,e0)
          (cons (set e0) (store-extend store x (set-union (set x) (lookup args))))]
         [`(let ([,x (prim make-vector ,a ,b)]) ,e0)
          (cons (set e0) (store-extend store x (set-union (set x) (lookup a) (lookup b))))]
         [`(let ([,x (apply-prim make-vector ,args)]) ,e0)
          (cons (set e0) (store-extend store x (set-union (set x) (lookup args))))]
         [`(let ([,x (prim vector-set! ,a ,b ,c)]) ,e0)
          (cons (set e0) (foldl (lambda (a store) (store-extend store a (lookup c)))
                                store
                                (filter symbol? (set->list (lookup a)))))]
         [`(let ([,x (apply-prim vector-set! ,a)]) ,e0)
          (cons (set e0) (foldl (lambda (a store) (store-extend store a (lookup a)))
                                store
                                (filter symbol? (set->list (lookup a)))))]
         [`(let ([,x (prim vector-ref ,a ,bs ...)]) ,e0)
          (cons (set e0) (store-extend store x (foldl (lambda (a vs)
                                                        (set-union vs (lookup a)))
                                                      (set)
                                                      (filter symbol? (set->list (lookup a))))))]
         [`(let ([,x (apply-prim vector-ref ,args)]) ,e0)
          (cons (set e0) (store-extend store x (foldl (lambda (a vs)
                                                        (set-union vs (lookup a)))
                                                      (set)
                                                      (filter symbol? (set->list (lookup args))))))]
         [`(let ([,x (prim ,op ,as ...)]) ,e0)
          (cons (set e0) (store-extend store x (foldl set-union (set) (map lookup as))))]
         [`(let ([,x (apply-prim ,op ,a)]) ,e0)
          (cons (set e0) (store-extend store x (lookup a)))]
         [`(if ,ae ,e0 ,e1)
          (cons (set e0 e1) store)]
         [`(apply ,f ,a)
          (define vs (lookup a))
          (foldl (lambda (lam succs+store)
                   (match-define (cons succs store) succs+store)
                   (match lam
                          [`(lambda (,xs ...) ,e0)
                           (cons (set-add succs e0)
                                 (foldl (lambda (x store)
                                          (store-extend store x vs))
                                        store
                                        xs))]
                          [`(lambda ,(? symbol? x) ,e0)
                           (cons (set-add succs e0)
                                 (store-extend store x vs))]
                          [else succs+store]))
                 (cons (set) store)
                 (set->list (lookup f)))]
         [`(,f ,as ...) 
          (foldl (lambda (lam succs+store)
                   (match-define (cons succs store) succs+store)
                   (match lam
                          [`(lambda (,xs ...) ,e0)
                           #:when (= (length xs) (length as))
                           (cons (set-add succs e0)
                                 (foldl (lambda (x a store)
                                          (store-extend store x (lookup a)))
                                        store
                                        xs
                                        as))]
                          [`(lambda ,(? symbol? x) ,e0)
                           (cons (set-add succs e0)
                                 (foldl (lambda (a store)
                                          (store-extend store x (lookup a)))
                                        store
                                        as))]
                          [else succs+store]))
                 (cons (set) store)
                 (set->list (lookup f)))]))


; utility functions
(define (graph-extend graph e0 succs)
  (foldl (lambda (succ graph)
           (graph-extend graph succ (set)))
         (hash-set graph e0 (set-union succs (hash-ref graph e0 set)))
         (set->list succs)))
(define (store-extend store x val)
  (hash-set store x (set-union (store-ref store x) (if (set? val) val (set val)))))
(define (store-ref store x)
  (hash-ref store x set))



; Run 0-CFA analysis (times out and returns (cons #f #f) if too expensive)
(define (0-cfa exp)
  (define start-ms (current-milliseconds))
  (let loop ([graph (hash exp (set))]
             [store (hash)])
    (match-define (cons graph+ store+)
                  ; fold over all reachable expressions, process them to update
                  ; the store and obtain a set of potentially newly reached expressions
                  (foldl (lambda (e0 graph+store)
                           (match-define (cons graph store) graph+store)
                           (match-define (cons succs store+)
                                         (successors e0 store))
                           (cons (graph-extend graph e0 succs)
                                 store+))
                         (cons graph store)
                         (hash-keys graph)))
    ; If graph and store are at a fixed point, return them, otherwise loop
    (if (and (equal? graph graph+)
             (equal? store store+))
        (cons graph+ store+)
        (if (> (- (current-milliseconds) start-ms) 19000)
            (begin #;(pretty-print '0-cfa-timed-out)
                   (cons #f #f))
            (loop graph+ store+)))))






